﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace HW09CPC
{
    /// Nehmen Formel fuer Schwarzkoerperstrahlung, nehmen Relation dreier Wellenlaengen
    /// und generieren daraus eine einfache Colormap die von Schwarz ueber rot nach hellblau geht.
    /// Kopiert von http://de.wikipedia.org/wiki/Wiensches_Verschiebungsgesetz
    class ColormapHeat
    {
        private const double H = 6.62606957E-34;
        private const double C = 299792458;
        private const double K = 1.3806488E-23;

        private const double LAMBDA_R = 680E-9;
        private const double LAMBDA_G = 550E-9;
        private const double LAMBDA_B = 450E-9;

        private const double T_MIN = 0;
        private const double T_MAX = 8000;
        private const double T_SINUS = 2000;

        private int m_steps;
        private float[] m_map = null;

        /// Erstellt eine neue Colormap mit der angegebenen Anzahl an Stützpunkten
        /// <param name="steps">Größe des Arrays</param>
        public ColormapHeat(int steps)
        {
            if (steps < 2)
                throw new ArgumentException("Muss mind. zwei Werte haben");

            m_steps = steps;
        }

        /// Generiert das gewünschte Array beim ersten Aufruf, liefert das zwischengespeicherte Array bei jedem Folgeaufruf zurück.
        /// <returns>Array mit der im Konstruktor spezifizierten Länte</returns>
        public float[] getRgbaMap()
        {
            if (m_map != null)
                return null;

            m_map = new float[4 * m_steps];
            
            double stepsize = (T_MAX - T_MIN) / (m_steps - 1);

            for (int i = 0; i < m_steps; i++)
            {
                int cpos = 4 * i;
                double T = T_MIN + i * stepsize;

                double r = wienvg(T, LAMBDA_R);
                double g = wienvg(T, LAMBDA_G);
                double b = wienvg(T, LAMBDA_B);

                double maxv = Math.Max(Math.Max(r, g), b);
                if (maxv > 0)
                {
                    r /= maxv;
                    g /= maxv;
                    b /= maxv;
                }

                if (T < T_SINUS)
                {
                    /* Einschwingen von Schwarz */
                    double scal = Math.Sin((T / T_SINUS) * 2 * Math.PI / 4);
                    r *= scal;
                    g *= scal;
                    b *= scal;

                }

                m_map[cpos + 0] = (float)r;
                m_map[cpos + 1] = (float)g;
                m_map[cpos + 2] = (float)b;
                m_map[cpos + 3] = 1;
            }


            return m_map;
        }

        private double wienvg(double T, double lambda)
        {
            /* Kopiert von http://de.wikipedia.org/wiki/Wiensches_Verschiebungsgesetz */
            return ( (2*Math.PI*H*Math.Pow(C,2)) / Math.Pow(lambda,5) ) / (Math.Exp((H*C)/(lambda*K*T)) - 1);
        }

    }
}
