﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace HW09CPC
{
    /// Stellt eine Colormap dar, die von Schwarz nach Rot über Gelb nach Grün und anschließend über Türkis nach Blau marschiert
    class ColormapHsv
    {
        private int m_steps;
        private float[] m_map = null;
        private const double SINUS_INT = 0.1;

        /// Erstellt eine neue Colormap mit der angegebenen Anzahl an Stützpunkten
        /// <param name="steps">Größe des Arrays</param>
        public ColormapHsv(int steps)
        {
            if (steps < 2)
                throw new ArgumentException("Mindestens zwei Werte notwendig");
            m_steps = steps;
        }

        /// Generiert das gewünschte Array beim ersten Aufruf, liefert das zwischengespeicherte Array bei jedem Folgeaufruf zurück.
        /// <returns>Array mit der im Konstruktor spezifizierten Länte</returns>
        public float[] getRgbaMap()
        {
            if (m_map != null)
                return m_map;

            m_map = new float[4 * m_steps];

            for (int i = 0; i < m_steps; i++)
            {
                int cpos = 4 * i;
                double zopos = (double)i / (m_steps - 1);
                double scale = sinusoid(zopos);

                m_map[cpos + 0] = (float)(scale*calcR(zopos));
                m_map[cpos + 1] = (float)(scale*calcG(zopos));
                m_map[cpos + 2] = (float)(scale*calcB(zopos));
                m_map[cpos + 3] = 1;
            }

            return m_map;
        }

        /* Formeln nach Stift und Papier */
        private double calcR(double x)
        {
            if (x <= 0.25) return 1;
            if (x <= 0.5) return 1 - (x - 0.25)*4;
            return 0;
        }

        private double calcG(double x)
        {
            if (x <= 0) return 0;
            if (x <= 0.25) return 4 * x;
            if (x <= 0.75) return 1;
            if (x <= 1) return 1 - (x - 0.75)*4;
            return 0;
        }

        private double calcB(double x)
        {
            if (x <= 0.5) return 0;
            if (x <= 0.75) return (x - 0.5)*4;
            return 1;
        }

        private double sinusoid(double x)
        {
            if (x <= 0) return 0;
            if (x < SINUS_INT) return Math.Sin((x / SINUS_INT) * 2 * Math.PI / 4);
            else return 1;
        }
    }
}
