﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using OpenTK.Graphics.OpenGL;
using System.Drawing.Imaging;

namespace HW09CPC
{
    public partial class Form1 : Form
    {
        // Raw/Processed data
        private PcCvsDataset mCvsDataset = null;
        private readonly object mCvsDatasetLock = new object();
        private int[] mLastdimsel = null;

        // Screen quad vars
        private PrimitiveDataInfo mScreenQuadUvInfo;
        private PrimitiveData mScreenQuadUvVbo;
        private GLProgram mScreenQuadTex2dProg;
        
        // Basic PC axis geom. data
        private PrimitiveDataInfo mBasicPcInfo;
        private List<PrimitiveData> mBasicPcVbos;
        private GLProgram mBasicPcPlotProg;
        private bool mUseLocalTrans = false;

        // PC Density Plot Vars
        private GLTextureRgba2D mDensityPcPlotTexture;
        private GLFbo2D mDensityPcPlotFbo;

        // Color mapping vars
        private const float COLORMAPPING_LINE_SCALE = 0.01f;
        private const uint COLORMAP_RESOLUTION = 100;
        private float mColorMappingLineWidth = 5;
        private GLTextureRgba1D mGreyTexture;
        private GLTextureRgba1D mHeatTexture;
        private GLTextureRgba1D mRainbowTexture;
        private GLTextureRgba1D mActiveColormap = null;
        private GLProgram mColorMappingProg;

        // Density texture vars
        private GLProgram mDensityTextureProg;
        private uint mDensityTextureResolution;
        private List<GLTextureRgba2D> mDensityTextures;
        private List<GLFbo2D> mDensityTextureFbos;
        private PrimitiveDataInfo mDensityTextureInfo;
        private List<PrimitiveData> mDensityTextureVbos;
        private bool mUseBilinearSigma = true;

        // Continuous pc plot vars
        private const float CONTINUOUS_PCPLOT_LINE_SCALE = 0.01f;
        private float mContinuousPcPlotLineWidth = 8;
        private int mDensityTexSampleCount;
        private PrimitiveDataInfo mContinuousPcPlotInfo;
        private PrimitiveData mContinuousPcPlotGoemetryInvokeVbo;
        private GLTextureRgba2D mContinuousPcPlotTexture;
        private GLFbo2D mContinuousPcPlotFbo;
        private GLProgram mContinuousPcPlotProg;
        private GLTextureRgba2D mContinuousPcPlotNoiseTexture;
        private const int NOISE_TEX_SIZE = 256;

        // MIPMAP VARS
        private MipmapGenerator mDensityMipmapGenerator;
        private MipmapGenerator mContinuousMipmapGenerator;
        private GLProgram mMipmapGeneratorProg;

        // AXIS VARS
        private GLProgram mAxisProg;
        
        // Auto resize FBOs
        private List<GLFbo2D> mAutoResizeFbos;
        
        private bool m_iscontrolloaded = false;

        // GUI selection stuff
        private ucSteering.RenderModeEnum mCurrentRendermode = ucSteering.RenderModeEnum.Simple;

        /// Konstruktor
        public Form1()
        {
            InitializeComponent();
        }
        
        private void Form1_Load(object sender, EventArgs e)
        {
            // hen the Load event of the form is triggered. Don't touch glControl/GL
        }

        private void glControl1_Load(object sender, EventArgs e)
        {
            // Then the Load event of the GLControl is triggered. OK to touch glControl/GL
            ucSteering1.DimensionOrderChanged += new ucSteering.DimensionEventHandler(ucSteering1_DimensionOrderChanged);
            ucSteering1.RenderModeChanged += new ucSteering.RenderModeEventHandler(ucSteering1_RenderModeChanged);
            ucSteering1.ContinuousParsChanged += new ucSteering.ContinuousParsEventHandler(ucSteering1_ContinuousParsChanged);
            ucSteering1.ColormapChanged += new ucSteering.ColormapEventHandler(ucSteering1_ColormapChanged);
            ucSteering1.TransformParsChanged += new ucSteering.TransformParsEventHandler(ucSteering1_TransformParsChanged);

            m_iscontrolloaded = true;
            GL.ClearColor(0f,0f,0f,0f); // soll bei add. blending so sein (logisch denken)
            
            SetupViewPort();
            
            GL.Enable(EnableCap.Blend);
            GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.One);
            GL.Enable(EnableCap.LineSmooth);
            GL.Hint(HintTarget.LineSmoothHint, HintMode.Nicest);

            GL.PointSize(1.0f);
            
            // Auto resize FBOs
            mAutoResizeFbos = new List<GLFbo2D>();

            // BASIC PC VARS
            mBasicPcInfo = new PrimitiveDataInfo(BeginMode.Lines);
            mBasicPcInfo.appendAttribute("yValue", 1, 0);
            mBasicPcPlotProg = GLProgram.buildProgram("shaders\\ndcBasicPc.vert", null, "shaders\\singleColor.frag", mBasicPcInfo.getNames());
       
            // SCREEN QUAD VARS
            mScreenQuadUvInfo = new PrimitiveDataInfo(BeginMode.Triangles);
            mScreenQuadUvInfo.appendAttribute("inPosition", 2, 0);
            mScreenQuadUvInfo.appendAttribute("inTexCoord2d", 2, 2);
            mScreenQuadUvVbo = new PrimitiveData(Utility.getNdcQuadWithUV(), mScreenQuadUvInfo, BufferUsageHint.StaticDraw, true);
            mScreenQuadTex2dProg = GLProgram.buildProgram("shaders\\screenQuad.vert", null, "shaders\\single2dTex.frag", mScreenQuadUvInfo.getNames());
                        
            // PC Density Plot Vars
            mDensityPcPlotTexture = new GLTextureRgba2D("PcDensityPlot", 1, 1, TextureUnit.Texture0, false, false, true, null);
            mDensityPcPlotFbo = new GLFbo2D(mDensityPcPlotTexture);
            mDensityPcPlotFbo.resize((uint)glControl1.Width, (uint)glControl1.Height);
            mAutoResizeFbos.Add(mDensityPcPlotFbo);

            // Color mapping vars
            ColormapGray greyColMap = new ColormapGray((int)COLORMAP_RESOLUTION);
            mGreyTexture = new GLTextureRgba1D("greyColormap", COLORMAP_RESOLUTION, TextureUnit.Texture1, false, false, true, greyColMap.getRgbaMap());
            ColormapHeat heatColMap = new ColormapHeat((int)COLORMAP_RESOLUTION);
            mHeatTexture = new GLTextureRgba1D("heatColormap", COLORMAP_RESOLUTION, TextureUnit.Texture1, false, false, true, heatColMap.getRgbaMap());
            ColormapHsv rainbowColMap = new ColormapHsv((int)COLORMAP_RESOLUTION);
            mRainbowTexture = new GLTextureRgba1D("heatColormap", COLORMAP_RESOLUTION, TextureUnit.Texture1, false, false, true, rainbowColMap.getRgbaMap());
            mColorMappingProg = GLProgram.buildProgram("shaders\\screenQuad.vert", null, "shaders\\colorMapping.frag", mScreenQuadUvInfo.getNames());

            // DENSITY TEXTURE VARS
            mDensityTextureResolution = 16;         
            mDensityTextureInfo = new PrimitiveDataInfo(BeginMode.Points);
            mDensityTextureInfo.appendAttribute("inPosition", 2, 0);
            mDensityTextureProg = GLProgram.buildProgram("shaders\\densityTextureGen.vert", null, "shaders\\densityTextureGen.frag", mDensityTextureInfo.getNames());

            // Continuous pc plot vars
            mDensityTexSampleCount = 128;
            mContinuousPcPlotInfo = new PrimitiveDataInfo(BeginMode.Points);
            mContinuousPcPlotInfo.appendAttribute("inPosition", 1, 0);
            mContinuousPcPlotGoemetryInvokeVbo = new PrimitiveData(new float[mDensityTexSampleCount * mDensityTexSampleCount], mContinuousPcPlotInfo, BufferUsageHint.StaticDraw, true);
            mContinuousPcPlotTexture = new GLTextureRgba2D("densityPcPlot", 1, 1, TextureUnit.Texture0, false, false, true, null);
            mContinuousPcPlotFbo = new GLFbo2D(mContinuousPcPlotTexture);
            mContinuousPcPlotFbo.resize((uint)glControl1.Width, (uint)glControl1.Height);
            mAutoResizeFbos.Add(mContinuousPcPlotFbo);
            mContinuousPcPlotProg = GLProgram.buildProgram("shaders\\densityPcPlot.vert", "shaders\\densityPcPlot.geom", "shaders\\densityPcPlot.frag", mContinuousPcPlotInfo.getNames());
            // load noise texture
            float[] imgData = new float[NOISE_TEX_SIZE * NOISE_TEX_SIZE * 4];            
            Random r = new Random();
            int i = 0;
            for (int x = 0; x < NOISE_TEX_SIZE; x++)
            {
                for (int y = 0; y < NOISE_TEX_SIZE; y++)
                {
                        imgData[i] = (float)r.NextDouble(); i++;
                        imgData[i] = (float)r.NextDouble(); i++;
                        imgData[i] = (float)r.NextDouble(); i++;
                        imgData[i] = (float)r.NextDouble(); i++;
                }
            }
            mContinuousPcPlotNoiseTexture = new GLTextureRgba2D("noise", (uint)NOISE_TEX_SIZE, (uint)NOISE_TEX_SIZE, TextureUnit.Texture4, false, false, true, imgData);
            mContinuousPcPlotNoiseTexture.setNearestFiltering();
                        
            // MIPMAP VARS 
            mMipmapGeneratorProg= GLProgram.buildProgram("shaders\\screenQuad.vert", null, "shaders\\generateMipmap.frag", mScreenQuadUvInfo.getNames());
            mDensityMipmapGenerator = new MipmapGenerator(this, mDensityPcPlotTexture, mMipmapGeneratorProg);
            mContinuousMipmapGenerator = new MipmapGenerator(this, mContinuousPcPlotTexture, mMipmapGeneratorProg);

            // AXIS VARS
            mAxisProg = GLProgram.buildProgram("shaders\\densityPcPlot.vert", "shaders\\axisPlot.geom", "shaders\\densityPcPlot.frag", mContinuousPcPlotInfo.getNames());
        }

        private void ucSteering1_TransformParsChanged(object sender, ucSteering.TransformParsEventargs e)
        {
            mUseLocalTrans = e.getLocalTransform();
            glControl1.Invalidate();
        }

        private void ucSteering1_ColormapChanged(object sender, ucSteering.ColormapEventargs e)
        {
            switch(e.getRenderMode())
            {
                case ucSteering.ColormapEnum.Gray:
                    mActiveColormap = mGreyTexture;
                    break;
                case ucSteering.ColormapEnum.Heat:
                    mActiveColormap = mHeatTexture;
                    break;
                case ucSteering.ColormapEnum.Rainbow:
                    mActiveColormap = mRainbowTexture;
                    break;
                default:
                    throw new ArgumentException("Das ist keine Colormap");
            }
            glControl1.Invalidate();
        }

        private void ucSteering1_ContinuousParsChanged(object sender, ucSteering.ContinuousParsEventargs e)
        {
            mDensityTextureResolution = (uint)e.getSigmaWidth();
            mDensityTexSampleCount = e.getSamplingPrec();
            if (mLastdimsel != null)
            {
                initDensityTextures(mLastdimsel.Length);

                // cleanup
                if (mContinuousPcPlotGoemetryInvokeVbo != null)
                    mContinuousPcPlotGoemetryInvokeVbo.delete();

                mContinuousPcPlotGoemetryInvokeVbo = new PrimitiveData(new float[mDensityTexSampleCount * mDensityTexSampleCount], mContinuousPcPlotInfo, BufferUsageHint.StaticDraw, true);

                mUseBilinearSigma = e.getUseBilinear();
                foreach (GLFbo2D dtfbo in mDensityTextureFbos)
                {
                    if (mUseBilinearSigma)
                        dtfbo.getColorAttachment().setLinearFiltering();
                    else
                        dtfbo.getColorAttachment().setNearestFiltering();
                }
            }
            glControl1.Invalidate();
        }

        private void ucSteering1_RenderModeChanged(object sender, ucSteering.RenderModeEventargs e)
        {
            mCurrentRendermode = e.getRenderMode();
            glControl1.Invalidate();
            
        }

        private void ucSteering1_DimensionOrderChanged(object sender, EventArgs e)
        {
            glControl1.Invalidate();
        }

        private void glControl1_Paint(object sender, PaintEventArgs e)
        {
            if (!m_iscontrolloaded) // Play nice
                return;

            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            render();

            Utility.GLErrorCheck("before SwapBuffers()");

            glControl1.SwapBuffers();

            updateScaleLabels();
        }

        private void render()
        {
            lock (mCvsDatasetLock)
            {
                if (mCvsDataset != null)
                {
                    updatePcVbos();
                    switch (mCurrentRendermode)
                    {
                        case ucSteering.RenderModeEnum.Simple:
                            renderSimple();
                            break;

                        case ucSteering.RenderModeEnum.Colormap:
                            renderColorMapping();
                            break;

                        case ucSteering.RenderModeEnum.Continuous:
                            renderContinuous();
                            break;

                        default:
                            throw new ArgumentException();
                    } /* switch */

                    renderAxis();

                } /* else nichts zu zeichnen */
                
            } /* lock */

            //GLVideoMemoryQuery.displayVideoMemoryInfo();
        }

        private void renderAxis()
        {
            GL.LineWidth(4);
            GL.Disable(EnableCap.Blend);

            mAxisProg.bind();

            int axisCountULoc = mAxisProg.GetUniformLocation("axisCount");
            GL.Uniform1(axisCountULoc, mLastdimsel.Length);

            int axisColorULoc = mAxisProg.GetUniformLocation("axisColor");
            GL.Uniform4(axisColorULoc, 0f, 0f, 0f, 1f);
            
            mContinuousPcPlotGoemetryInvokeVbo.activate();
            mContinuousPcPlotGoemetryInvokeVbo.draw(0, mLastdimsel.Length);
            mContinuousPcPlotGoemetryInvokeVbo.disable();

            GL.LineWidth(2);

            GL.Uniform4(axisColorULoc, 1f, 1f, 1f, 1f);

            mContinuousPcPlotGoemetryInvokeVbo.activate();
            mContinuousPcPlotGoemetryInvokeVbo.draw(0, mLastdimsel.Length);
            mContinuousPcPlotGoemetryInvokeVbo.disable();

            mAxisProg.unbind();

            GL.Enable(EnableCap.Blend);
            GL.LineWidth(1);
        }

        private void renderSimple()
        {
            mBasicPcPlotProg.bind();

            /* setting for global scaling (local override later) */
            float gMin = mCvsDataset.getGlobalMinimum(mLastdimsel);
            float gMax = mCvsDataset.getGlobalMaximum(mLastdimsel);
            int y0MinULoc = mBasicPcPlotProg.GetUniformLocation("y0Min");
            int y0MaxULoc = mBasicPcPlotProg.GetUniformLocation("y0Max");
            int y1MinULoc = mBasicPcPlotProg.GetUniformLocation("y1Min");
            int y1MaxULoc = mBasicPcPlotProg.GetUniformLocation("y1Max");
            GL.Uniform1(y0MinULoc, gMin);
            GL.Uniform1(y0MaxULoc, gMax);
            GL.Uniform1(y1MinULoc, gMin);
            GL.Uniform1(y1MaxULoc, gMax);

            int axisCountULoc = mBasicPcPlotProg.GetUniformLocation("axisCount");
            GL.Uniform1(axisCountULoc, mLastdimsel.Length);

            int currentAxisIndexULoc = mBasicPcPlotProg.GetUniformLocation("currentAxisIndex");

            int curAxis = 0;
            foreach (PrimitiveData p in mBasicPcVbos)
            {
                GL.Uniform1(currentAxisIndexULoc, curAxis);

                if (mUseLocalTrans)
                {
                    GL.Uniform1(y0MinULoc, mCvsDataset.getMinimum(mLastdimsel[curAxis]));
                    GL.Uniform1(y0MaxULoc, mCvsDataset.getMaximum(mLastdimsel[curAxis]));
                    GL.Uniform1(y1MinULoc, mCvsDataset.getMinimum(mLastdimsel[curAxis+1]));
                    GL.Uniform1(y1MaxULoc, mCvsDataset.getMaximum(mLastdimsel[curAxis+1]));
                }

                p.activate();
                p.draw();
                p.disable();

                curAxis++;
            }

            mBasicPcPlotProg.unbind();
        }

        private void renderColorMapping()
        {
            GL.LineWidth(mColorMappingLineWidth);
            
            // render simple in FBO
            mDensityPcPlotFbo.bindClear();
            renderSimple();
            mDensityPcPlotFbo.unbind();

            applyColorMapping(mDensityPcPlotTexture, mDensityMipmapGenerator); 

            GL.LineWidth(1);
        }

        private void applyColorMapping(GLTextureRgba2D _tex, MipmapGenerator _mmg)
        {
            _mmg.calcMipmaps();

            //GLTextureRgba1D colormap = mRainbowTexture;
            if (mActiveColormap == null)
                mActiveColormap = mGreyTexture;

            // do colormapping into backbuffer
            mColorMappingProg.bind();

            // Quad transform
            int stULoc = mColorMappingProg.GetUniformLocation("scaleTrans");
            float[] st = new float[4];
            st[0] = 1f; st[1] = 1f; // scale
            st[2] = 0f; st[3] = 0f; // translation
            GL.Uniform4(stULoc, 1, st);

            // Quad texture
            _tex.bind();
            int texULoc = mColorMappingProg.GetUniformLocation("tex");
            GL.Uniform1(texULoc, _tex.getTextureUnit() - TextureUnit.Texture0);

            // Colormap texture
            mActiveColormap.bind();
            int colormapULoc = mColorMappingProg.GetUniformLocation("colormap");
            GL.Uniform1(colormapULoc, mActiveColormap.getTextureUnit() - TextureUnit.Texture0);

            // max texture
            GLTextureRgba2D maxTex = _mmg.getMaxTex();
            
            maxTex.bind();

            int maxTexULoc = mColorMappingProg.GetUniformLocation("maxTex");
            GL.Uniform1(maxTexULoc, maxTex.getTextureUnit() - TextureUnit.Texture0);

            float maxLevel = (float)_mmg.getMaxLevel();
            int maxLevelULoc = mColorMappingProg.GetUniformLocation("maxLevel");
            GL.Uniform1(maxLevelULoc, maxLevel);

            mScreenQuadUvVbo.activate();
            mScreenQuadUvVbo.draw();
            mScreenQuadUvVbo.disable();

            mActiveColormap.unbind();
            maxTex.unbind();
            _tex.unbind();

            mColorMappingProg.unbind();
        }

        private void renderContinuous()
        {            
            renderDensityTextures();
            renderContinuousPcPlot();                        
            applyColorMapping(mContinuousPcPlotTexture, mContinuousMipmapGenerator);
        }

        private void renderDensityTextures()
        {
            mDensityTextureProg.bind();
            /* global parameter first, local override later */
            int min0ULoc = mDensityTextureProg.GetUniformLocation("minimum0");
            int max0ULoc = mDensityTextureProg.GetUniformLocation("maximum0");
            int min1ULoc = mDensityTextureProg.GetUniformLocation("minimum1");
            int max1ULoc = mDensityTextureProg.GetUniformLocation("maximum1");
            float gmin = mCvsDataset.getGlobalMinimum(mLastdimsel);
            float gmax = mCvsDataset.getGlobalMaximum(mLastdimsel);
            GL.Uniform1(min0ULoc, gmin);
            GL.Uniform1(max0ULoc, gmax);
            GL.Uniform1(min1ULoc, gmin);
            GL.Uniform1(max1ULoc, gmax);

            int i = 0;
            foreach (GLFbo2D fbo in mDensityTextureFbos)
            {
                PrimitiveData vao = mDensityTextureVbos.ElementAt<PrimitiveData>(i);

                if (mUseLocalTrans)
                {
                    GL.Uniform1(min0ULoc, mCvsDataset.getMinimum(mLastdimsel[i]));
                    GL.Uniform1(max0ULoc, mCvsDataset.getMaximum(mLastdimsel[i]));
                    GL.Uniform1(min1ULoc, mCvsDataset.getMinimum(mLastdimsel[i+1]));
                    GL.Uniform1(max1ULoc, mCvsDataset.getMaximum(mLastdimsel[i+1]));
                }

                fbo.bindClear();
                vao.activate();
                vao.draw();
                vao.disable();               
                fbo.unbind();

                i++;
            }

            mDensityTextureProg.unbind();
        }
        
        private void renderContinuousPcPlot()
        {
            GL.LineWidth(mContinuousPcPlotLineWidth);

            // bind everything
            mContinuousPcPlotFbo.bindClear();
            mContinuousPcPlotProg.bind();
            
            // init uniforms

            int sampleCountULoc = mContinuousPcPlotProg.GetUniformLocation("sampleCount");
            GL.Uniform1(sampleCountULoc, mDensityTexSampleCount);
            
            int axisCountULoc = mContinuousPcPlotProg.GetUniformLocation("axisCount");
            GL.Uniform1(axisCountULoc, mLastdimsel.Length);

            mContinuousPcPlotNoiseTexture.bind();            
            int noiseTexULoc = mContinuousPcPlotProg.GetUniformLocation("noiseTex");
            GL.Uniform1(noiseTexULoc, mContinuousPcPlotNoiseTexture.getTextureUnit() - TextureUnit.Texture0);

            int i = 0;

            foreach (GLFbo2D fbo in mDensityTextureFbos)
            {
                // bind density texture            
                GLTextureRgba2D densityTex = fbo.getColorAttachment();
                densityTex.bind();

                int densityTexULoc = mContinuousPcPlotProg.GetUniformLocation("densityTex");
                GL.Uniform1(densityTexULoc, densityTex.getTextureUnit() - TextureUnit.Texture0);
                                                
                int currentAxisIndexULoc = mContinuousPcPlotProg.GetUniformLocation("currentAxisIndex");
                GL.Uniform1(currentAxisIndexULoc, i);

                // draw everything
                mContinuousPcPlotGoemetryInvokeVbo.activate();
                mContinuousPcPlotGoemetryInvokeVbo.draw();
                mContinuousPcPlotGoemetryInvokeVbo.disable();

                // unbind everything
                densityTex.unbind();

                i++;
            }

            mContinuousPcPlotNoiseTexture.unbind();

            mContinuousPcPlotProg.unbind();
            mContinuousPcPlotFbo.unbind();
            GL.LineWidth(1);
        }

        private void updatePcVbos()
        {
            int[] dimsel = ucSteering1.getOrderedDimSelection();
            if (mLastdimsel == null || !Enumerable.SequenceEqual(mLastdimsel, dimsel))
            {
                /* neuer Datensatz oder dimensionsauswahl hat sich geaendert muss daten neu laden */
                
                // cleanup
                if (mBasicPcVbos != null)
                {
                    foreach (PrimitiveData vbo in mBasicPcVbos)
                        vbo.delete();

                    mBasicPcVbos.Clear();
                }

                mBasicPcVbos = new List<PrimitiveData>();

                for (int i = 0; i < dimsel.Length - 1; i++)
                {
                    float[] points = mCvsDataset.create2dProjDataDomain(dimsel[i], dimsel[i + 1]);
                    mBasicPcVbos.Add(new PrimitiveData(points, mBasicPcInfo, BufferUsageHint.StaticDraw, true));
                }

                initDensityTextures(dimsel.Length);
                
                mLastdimsel = dimsel;
            }            
        }

        private void initDensityTextures(int nbrOfTexs)
        { 
            // cleanup
            if (mDensityTextureFbos!=null)
            {
                foreach (GLFbo2D fbo in mDensityTextureFbos)
                    fbo.delete();

                mDensityTextureFbos.Clear();
            }

            if (mDensityTextures != null)
            {
                foreach (GLTextureRgba2D tex in mDensityTextures)
                    tex.deleteTexture();

                mDensityTextures.Clear();
            }

            if (mDensityTextureVbos!=null)
            {
                foreach (PrimitiveData vbo in mDensityTextureVbos)
                    vbo.delete(false);

                mDensityTextureVbos.Clear();
            }

            // DENSITY TEXTURE VARS
            mDensityTextures = new List<GLTextureRgba2D>();
            mDensityTextureFbos = new List<GLFbo2D>();
            mDensityTextureVbos = new List<PrimitiveData>();

            for (int i = 0; i < nbrOfTexs - 1; i++)
            {
                // DENSITY TEXTURE VARS
                // FBOs
                GLTextureRgba2D dt = new GLTextureRgba2D("dt"+i, mDensityTextureResolution, mDensityTextureResolution, TextureUnit.Texture1, false, false, true, null);
                mDensityTextures.Add(dt);
                
                if(!mUseBilinearSigma) /* default ist linear */
                    dt.setNearestFiltering();

                mDensityTextureFbos.Add(new GLFbo2D(dt));

                // VBOs
                mDensityTextureVbos.Add(new PrimitiveData(mBasicPcVbos.ElementAt<PrimitiveData>(i).getVertexVbo(), mDensityTextureInfo, true));
            }
        }

        private void SetupViewPort()
        {
            int w = glControl1.Width;
            int h = glControl1.Height;
            GL.Viewport(0, 0, w, h); // Use all of the glControl painting area
        }

        private void updateLineWidth()
        {
            int h = glControl1.Height;

            mColorMappingLineWidth = Math.Max(1f, h*COLORMAPPING_LINE_SCALE);
            mContinuousPcPlotLineWidth = Math.Max(1f, h*CONTINUOUS_PCPLOT_LINE_SCALE);
        }

        private void glControl1_Resize(object sender, EventArgs e)
        {
            SetupViewPort();
            updateLineWidth();

            // RESIZE FBOS
            if (mAutoResizeFbos != null)
            {
                foreach (GLFbo2D fbo in mAutoResizeFbos)
                    fbo.resize((uint)glControl1.Width, (uint)glControl1.Height);
            }

            if (mDensityMipmapGenerator != null) mDensityMipmapGenerator.resize();
            if (mContinuousMipmapGenerator != null) mContinuousMipmapGenerator.resize();

            glControl1.Invalidate();          
        }

        private void mitLoadDs_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Multiselect = false;
            ofd.InitialDirectory = Directory.GetCurrentDirectory();
            ofd.Filter = "Comma separate values (*.csv)|*.csv|Alle files (*.*)|*.*";
            
            ofd.ShowDialog();

            if (String.IsNullOrEmpty(ofd.FileName)) return;

            lock (mCvsDatasetLock)
            {
                mCvsDataset = new PcCvsDataset(ofd.FileName);
                ucSteering1.setDimensions(mCvsDataset);
                mLastdimsel = null;
            }
            glControl1.Invalidate();
        }

        private void miQuit_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        /// Zeichnet ein texturiertes Rechteck (NDC) über den Framebuffer
        /// <param name="_tex">die Textur welche auf das Rechteck gemapped wird</param>
        /// <param name="_xScale">die x Skalierung des Rechtecks</param>
        /// <param name="_yScale">die y Skalierung des Rechtecks</param>
        /// <param name="_xTrans">die x Translation des Rechtecks</param>
        /// <param name="_yTrans">die y Translation des Rechtecks</param>
        /// <param name="_level">die zu verwendende LOD Stufe</param>
        public void drawScreenQuad(GLTextureRgba2D _tex, float _xScale = 1.0f, float _yScale = 1.0f, float _xTrans = 0.0f, float _yTrans = 0.0f, int _level = 0)
        {
            mScreenQuadTex2dProg.bind();

            // Quad transform
            int stULoc = mScreenQuadTex2dProg.GetUniformLocation("scaleTrans");
            float[] st = new float[4];
            st[0] = _xScale; st[1] = _yScale; // scale
            st[2] = _xTrans; st[3] = _yTrans; // translation
            GL.Uniform4(stULoc, 1, st);

            // Quad texture
            _tex.bind();
            int texULoc = mScreenQuadTex2dProg.GetUniformLocation("tex");
            GL.Uniform1(texULoc, _tex.getTextureUnit() - TextureUnit.Texture0);

            int levelULoc = mScreenQuadTex2dProg.GetUniformLocation("level");
            GL.Uniform1(levelULoc, _level);

            drawScreenQuadGeometry();

            _tex.unbind();
            mScreenQuadTex2dProg.unbind();
        }

        /// Aktiviert das VAO des Rechtecks und setzt den Zeichen-Befehl ab.
        public void drawScreenQuadGeometry()
        {       
            mScreenQuadUvVbo.activate();
            mScreenQuadUvVbo.draw();
            mScreenQuadUvVbo.disable();
        }

        private List<Label> mLabelsUpper = new List<Label>();
        private List<Label> mLabelsLower = new List<Label>();
        private void updateScaleLabels()
        {
            lock (mCvsDatasetLock)
            {

                if (mCvsDataset == null) return;

                int gcx = glControl1.Location.X;
                int gcy = glControl1.Location.Y;
                int gcw = glControl1.Width;
                int gch = glControl1.Height;

                int vpadding = 0;

                int labelw = 70;
                int labelh = 20;
                
                foreach (Label l in mLabelsUpper)
                    this.Controls.Remove(l);
                foreach (Label l in mLabelsLower)
                    this.Controls.Remove(l);
                mLabelsUpper.Clear();
                mLabelsLower.Clear();

                if (mLastdimsel.Length < 2) return;

                for (int i = 0; i < mLastdimsel.Length; i++)
                {
                    /* upper label */
                    Label ulabel = new Label();
                    ulabel.Width = labelw;
                    ulabel.Height = labelh;
                    ulabel.Text = mCvsDataset.getMaximum(mLastdimsel[i]).ToString("0.00");
                    ulabel.TextAlign = ContentAlignment.BottomCenter;
                    ulabel.Location = new Point(gcx - labelw / 2 + (gcw * i) / (mLastdimsel.Length - 1), gcy - labelh);
                    mLabelsUpper.Add(ulabel);
                    this.Controls.Add(ulabel);

                    /* lower label */
                    Label llabel = new Label();
                    llabel.Width = labelw;
                    llabel.Height = labelh;
                    llabel.Text = mCvsDataset.getMinimum(mLastdimsel[i]).ToString("0.00");
                    llabel.TextAlign = ContentAlignment.TopCenter;
                    llabel.Location = new Point(gcx - labelw / 2 + (gcw * i) / (mLastdimsel.Length - 1), gcy + gch);
                    mLabelsLower.Add(llabel);
                    this.Controls.Add(llabel);

                }
            } /* lock */
        }/* updateScaleLabels */

    }
}