﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Stellt die Kapselungen eines GL-Program dar.
    class GLProgram
    {
        private int mHandle;	
	    private GLShader mVertexShader;
	    private GLShader mGeometryShader;
	    private GLShader mFragmentShader;

        /// Konstruktor
        /// <param name="_vertexShader">der Vertex-Shader</param>
        /// <param name="_geometryShader">der Geometry-Shader</param>
        /// <param name="_fragmentShader">der Fragment-Shader</param>
        /// <param name="_attributes">die Attribut-Namen</param>
        public GLProgram(GLShader _vertexShader, GLShader _geometryShader, GLShader _fragmentShader, string[] _attributes)
        {
            mVertexShader = _vertexShader;
            mGeometryShader = _geometryShader;
            mFragmentShader = _fragmentShader;

	        if(mVertexShader == null)
		        throw new Exception("GLProgram - vertex shader invalid");
	
	        if(mFragmentShader == null)
		        throw new Exception("GLProgram - fragment shader invalid");
	
	        mHandle = GL.CreateProgram();	
				
	        GL.AttachShader(mHandle, mVertexShader.getHandle());
	        if(mGeometryShader != null)
		        GL.AttachShader(mHandle, mGeometryShader.getHandle());
	        GL.AttachShader(mHandle, mFragmentShader.getHandle());

            for (int i = 0; i < _attributes.Length; i++)
                GL.BindAttribLocation(mHandle, i, _attributes[i]);
	
	        link();
        }

        /// Löscht das GL-Progam
        public void delete()
        {
            GL.DeleteProgram(mHandle);
        }

        private void link()
        {
	        GL.LinkProgram(mHandle);

	        int linked;
	        GL.GetProgram(mHandle, ProgramParameter.LinkStatus, out linked);
	
	        if (linked == 0)
	        {	        
                Console.Out.WriteLine("Program linking failed - Error log:");
                Console.Out.Write(GL.GetProgramInfoLog(mHandle));
                		
		        throw new Exception("GLProgram - program linking failed");
	        }
        }

        /// Aktiviert das GL-Progam
        public void bind()
        {
	        GL.UseProgram(mHandle);
        }

        /// Deaktiviert das GL-Progam
        public void unbind()
        {
	        GL.UseProgram(0);
        }

        /// Liefert die ID des uniformen Attributes mittels des Namens zurück 
        /// <param name="_name">der Name</param>
        /// <returns>die ID</returns>
        public int GetUniformLocation(string _name)
        {
	        int id = GL.GetUniformLocation(mHandle, _name);
	
	        if(id == -1)
		        Console.Out.WriteLine("Uniform location '" + _name + "' not found");

	        return id;
        }

        /// Eine Helper-Methode zur Erzeugung eines GLPrograms
        /// <param name="_vertexShaderPath">der Vertex-Shader Pfad</param>
        /// <param name="_geometryShaderPath">der Geometry-Shader Pfad</param>
        /// <param name="_fragmentShaderPath">der Fragment-Shader Pfad</param>
        /// <param name="_attributes">die Attribut-Namen</param>
        /// <returns>das erstellte GLProgram</returns>
        public static GLProgram buildProgram(string _vertexShaderPath, string _geometryShaderPath, string _fragmentShaderPath, string[] _attributes)
        {
            GLShader gs = null;
            if (_geometryShaderPath != null)
                gs = new GLShader(ShaderType.GeometryShader, _geometryShaderPath);

            GLShader vs = new GLShader(ShaderType.VertexShader, _vertexShaderPath);
            GLShader fs = new GLShader(ShaderType.FragmentShader, _fragmentShaderPath);

            return new GLProgram(vs, gs, fs, _attributes);
        }
    }
}
