﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Stellt die Kapselungen einer 2D RGBA-Textur dar.
    public class GLTextureRgba2D : GLTextureRgba
    {
        /// Konstruktor
        /// <param name="_name">der Name der Textur</param>
        /// <param name="_width">die Breite</param>
        /// <param name="_height">die Höhe</param>
        /// <param name="_textureUnit">die Textur-Einheit</param>
        /// <param name="_generateMipmap">true: Mipmaps werden generiert</param>
        /// <param name="_srgb">true: die Textur wird in SRGB angelegt</param>
        /// <param name="_generateTexture">true: die Textur wird sofort geladen</param>
        /// <param name="_data">das Array der Pixel-Daten</param>
        public GLTextureRgba2D(string _name, uint _width, uint _height, TextureUnit _textureUnit, bool _generateMipmap, bool _srgb, bool _generateTexture, float[] _data)
            : base(_name, _width, _height, _textureUnit, _generateMipmap, _srgb, _data)
        {
            if (_data != null)
            {
                if (mWidth * mHeight != _data.Length / 4)
                    throw new Exception("GLTextureRgba2D: data/dimentions mismatch");
            }

	        initMembers();

	        if(_generateTexture)
		        loadTexture();
        }

        private void initMembers()
        {
	        mTarget = TextureTarget.Texture2D;
        }

        /// Läd die Textur in den Grafikspeicher und das angegebene Mipmap-Level
        /// <param name="_level">das gewünschte Mipmap-Level</param>
        override public void uploadTexture(int _level = 0)
        {
            int norm = (int)Math.Pow(2.0, (double)_level);
            int levelWidth = Math.Max(1, (int)mWidth / norm);
            int levelHeight = Math.Max(1, (int)mHeight / norm);

            if(mData != null)
                GL.TexImage2D(mTarget, 0, mInternalFormat, (int)mWidth, (int)mHeight, 0, mFormat, mType, mData);
            else
                GL.TexImage2D(mTarget, _level, mInternalFormat, levelWidth, levelHeight, 0, mFormat, mType, mData);

            if (mIsMipmapped)
                generateMipmap();
        }
    }
}
