﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Stellt die Kapselungen eines GL-VBOs dar.
    class GLVbo
    {       
        private int mHandle;
        private BufferTarget mTarget;
        private BufferUsageHint mUsage;
        private int mDataLength;

        /// Konstruktor
        /// <param name="_vertexData">die Vertex-Daten</param>
        /// <param name="_usage">die VBO-Usage</param>
        public GLVbo(float[] _vertexData, BufferUsageHint _usage)
        {	
	        GL.GenBuffers(1, out mHandle);
	        loadVertexData(_vertexData, _usage);
        }

        /// Liefert die Länge der VBO-Daten zurück
        /// <returns>die Daten-Länge</returns>
        public int getDataLength()
        {
            return mDataLength;
        }

        /// Liefert die VBO-Usage zurück
        /// <returns>die VBO-Usage</returns>
        public BufferUsageHint getUsage()
        {
            return mUsage;
        }

        /// Löscht das VBO
        public void delete()
        {
	        GL.DeleteBuffers(1, ref mHandle);
        }

        /// Bindet das VBO
        public void bind()
        {
	        GL.BindBuffer(mTarget, mHandle);
        }

        /// Löst das VBO
        public void unbind()
        {
	        GL.BindBuffer(mTarget, 0);
        }

        /// Läd das VBO
        /// <param name="_vertexData">die Vertex-Daten</param>
        /// <param name="_usage">die VBO-Usage</param>
        public void loadVertexData(float[] _vertexData, BufferUsageHint _usage)
        {
            mDataLength = _vertexData.Length;
	        mTarget = BufferTarget.ArrayBuffer;
	        mUsage = _usage;
	
	        bind();
	        GL.BufferData(mTarget, (IntPtr) (_vertexData.Length*sizeof(float)), _vertexData, mUsage);
	        unbind();
        }

        /// Liefert die Art des VBOs zurück
        /// <returns>die VBO Art</returns>
        public BufferTarget getTarget()
        {
	        return mTarget;
        }
    }
}
