﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Helper-Klasse um Grafikspeicher-Daten auszulesen
    static class GLVideoMemoryQuery
    {
        // video memory query, code from http://www.geeks3d.com/20100531/programming-tips-how-to-know-the-graphics-memory-size-and-usage-in-opengl/
        private const int GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX = 0x9048;
        private const int GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX = 0x9049;
        private const string NV = "NVIDIA Corporation";

        private static string getVendor()
        {
            return GL.GetString(StringName.Vendor);
        }

        private static bool isNvidia()
        {
            return (getVendor().Equals(NV));
        }

        private static int getTotalVideoMemorySize()
        {
            int total_mem_kb = 0;
            if (isNvidia())
            {
                GL.GetInteger((GetPName)GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX, out total_mem_kb);
            }
            else
            {
                total_mem_kb = -1;
                System.Console.WriteLine("VideoMemoryQuery::getTotalVideoMemorySize ... Vendor not supported");
            }
            return total_mem_kb;
        }

        private static int getCurrentAvailableVideoMemorySize()
        {
            int cur_avail_mem_kb = 0;
            if (isNvidia())
            {
                GL.GetInteger((GetPName)GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX, out cur_avail_mem_kb);
            }
            else
            {
                cur_avail_mem_kb = -1;
                System.Console.WriteLine("VideoMemoryQuery::getCurrentAvailableVideoMemorySize ... Vendor not supported");
            }
            return cur_avail_mem_kb;
        }

        private static int getCurrentVideoMemoryUsage()
        {
            if (isNvidia())
            {
                return getTotalVideoMemorySize() - getCurrentAvailableVideoMemorySize();
            }
            else
            {
                System.Console.WriteLine("VideoMemoryQuery::getCurrentVideoMemoryUsage ... Vendor not supported");
                return -1;
            }
        }

        /// Gibt die Speicher-Daten in der Konsole aus.
        public static void displayVideoMemoryInfo()
        {
            if (isNvidia())
            {
                System.Console.WriteLine("- - - - - - - - - - - - -");
                System.Console.WriteLine("Video memory info (in KB)");
                System.Console.WriteLine("Total : " + getTotalVideoMemorySize());
                System.Console.WriteLine("Used  : " + getCurrentVideoMemoryUsage());
                System.Console.WriteLine("Free  : " + getCurrentAvailableVideoMemorySize());
                System.Console.WriteLine("- - - - - - - - - - - - -");
            }
            else
            {
                System.Console.WriteLine("VideoMemoryQuery::displayVideoMemoryInfo ... Vendor not supported");
            }
        }
    }
}
