﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace HW09CPC
{
    /**
     * Lädt CSV-Dateien und bietet parallelkoordinatenfreundlichen Zugriff auf deren Daten.
     * Sie soll außerdem Daten für eine passende Skalierung der Werte zur Verfügung stellen.
     */
    public class PcCvsDataset : IPcDataset
    {
        private static char[] P_SEP_CHARS = {'\t', ','};
        private const double RHO_OVERSHOOT = 1.0 / 16.0;

        private List<double[]> m_rmdat = new List<double[]>();
        private String[] m_dimtitles;
        private int m_ncols;
        private double[] m_ymaxima, m_yminima;
        private double m_yglobalmaximum, m_yglobalminimum;
        private float[][][] m_pcache;

        /// Erstellt einen neuen Datensatz anhand einer Datei 
        /// <param name="filename">Pfad zur Datei, die geladen werden soll</param>
        public PcCvsDataset(String filename)
        {
            /* Liest eine P_SEP_CHARS separierte CSV-Datei. 
             * Erste Zeile Betitelungen, erste Spalte Eintragslisten.
             * Doppelte Hochkomma werden getrimmt, Trennzeichen strikt nur n-mal pro Zeile*/
            StreamReader sr = new StreamReader(filename);

            try
            {

                String firstline, line;
                if (!sr.EndOfStream)
                    firstline = sr.ReadLine();
                else throw new IOException("Keine erste Zeile");

                int sint = int.MaxValue;
                for (int i = 0; i < P_SEP_CHARS.Length; i++)
                {
                    int p = firstline.IndexOf(P_SEP_CHARS[i]);
                    if (p != -1) sint = Math.Min(sint, p);
                }
                if (sint == int.MaxValue)
                    throw new IOException("Trennsymbol nicht gefunden");

                /* Erfasse Spaltentrennung */
                char sep_char = firstline[sint];
                String[] alltitles = firstline.Split(sep_char);
                m_ncols = alltitles.Length - 1;

                /* Extraktion Dimensionstitel */
                m_dimtitles = new String[m_ncols];
                Array.Copy(alltitles, 1, m_dimtitles, 0, m_ncols);
                for (int i = 0; i < m_dimtitles.Length; i++)
                    m_dimtitles[i] = m_dimtitles[i].Trim('"');

                while (!sr.EndOfStream)
                {
                    line = sr.ReadLine();
                    /* Aussteigen bei leerer Zeile */
                    if (string.IsNullOrEmpty(line))
                        break;

                    string[] linecols = line.Split(sep_char);
                    if (linecols.Length != m_ncols + 1)
                        throw new IOException("Zeilenlaenge nicht konstant");

                    m_rmdat.Add(new double[m_ncols]);
                    for (int i = 1; i < linecols.Length; i++)
                    {
                        try
                        {
                            double d = Double.Parse(linecols[i].Trim('"'));
                            m_rmdat.Last()[i - 1] = d;
                        }
                        catch (FormatException ex) 
                        {
                            throw new IOException("Das ist doch keine Zahl", ex);
                        }
                    }
                }

            }
            finally
            {
                sr.Close();
            }

            transformYTo01();

            /* Initialisiere leeren Cache */
            m_pcache = new float[m_ncols][][];
            for (int i = 0; i < m_ncols; i++)
                m_pcache[i] = new float[m_ncols][];
        }

        private void transformYTo01()
        {
            /* Daten row major. Y ist up vector */
            m_ymaxima = new double[m_ncols];
            m_yminima = new double[m_ncols];
            m_yglobalmaximum = double.NegativeInfinity;
            m_yglobalminimum = double.PositiveInfinity;

            for (int x = 0; x < m_ncols; x++)
            {
                double dmin = double.PositiveInfinity, dmax = double.NegativeInfinity;
                for (int y = 0; y < m_rmdat.Count; y++)
                {
                    double cval = m_rmdat[y][x];
                    dmin = Math.Min(dmin, cval);
                    dmax = Math.Max(dmax, cval);
                }
                m_ymaxima[x] = dmax + (dmax - dmin) * RHO_OVERSHOOT; /* All values should lie in [0,1) not [0,1] for density tex */
                m_yminima[x] = dmin - (dmax - dmin) * RHO_OVERSHOOT;
                m_yglobalmaximum = Math.Max(m_yglobalmaximum, m_ymaxima[x]);
                m_yglobalminimum = Math.Min(m_yglobalminimum, m_yminima[x]);

                ///* linear scale: (x-min)/(max-min) */
                //for (int y = 0; y < m_rmdat.Count; y++)
                //{
                //    if (dmax == dmin)
                //        m_rmdat[y][x] = 0.5;  /* vernünftig versagen  */
                //    else
                //        m_rmdat[y][x] = (m_rmdat[y][x] - dmin) / (dmax - dmin);
                //}
            }
        }

        /// Format: [x0, y0, r0, g0, b0, a0,  x1, y1, r1, g1, b1, a1,  ...,  xn, yn, rn, gn, bn, an] konkateniert fuer jedes line strip.
        /// Die Werte X sowie Y liegen in [0,1]. Diese Methode liefert ri=gi=bi=ai=1 für alle i.
        /// <returns>Erstelltes VBO</returns>
        [Obsolete("sollte keinen Grund geben nicht create2dProjDataDomain einzusetzen")]
        public float[] createDiscreteVbo()
        {
            float[] retval = new float[m_rmdat.Count * m_ncols * (2 + 4)];
            for (int i = 0; i < m_rmdat.Count; i++)
            {
                for (int nx = 0; nx < m_ncols; nx++)
                {
                    int soff = (i * m_ncols + nx) * (2 + 4);
                    float x = (float)((double)nx / (m_ncols - 1));
                    float y = (float)m_rmdat[i][nx];
                    retval[soff    ] = x;
                    retval[soff + 1] = y;
                    retval[soff + 2] = 1f; /* r */
                    retval[soff + 3] = 1f; /* g */
                    retval[soff + 4] = 1f; /* b */
                    retval[soff + 5] = 1f; /* a */

                }
            }
            return retval;
        }

        /// <summary>
        /// Liefert Orthogonalprojektion der Datenpunkte der Datendomäne an den gegebenen beiden Achsen.
        /// in Schleife mit Parametern (0,1), (1,2), (2,3), ..., (d-2, d-1). (d Dimension der Datendomäne)
        /// Format: [x0, y0, x1, y1, ..., xn, yn], jedes Element in [0,1].
        /// Da Datensatz statisch ist werden Abfragen gleich in einen Cache geschrieben.
        /// Diese Funktion ist bei Folgeaufrufe mit gleichen Parametern praktisch kostenlos.
        /// </summary>
        /// <param name="dim1">Index der ersten Dimension</param>
        /// <param name="dim2">Index der zweiten Dimension</param>
        /// <returns>Array im angegebenen Format</returns>
        public float[] create2dProjDataDomain(int dim1, int dim2)
        {
            if (dim1 < 0 || dim2 < 0)
                throw new ArgumentException("Mind. eine der beiden Dimensionen war kleiner als 0");
            if (dim1 >= m_ncols || dim2 >= m_ncols)
                throw new ArgumentException("Mind. eine der beiden Dimensionen hat (d-1) uebertroffen");

            if (m_pcache[dim1][dim2] != null)
                return m_pcache[dim1][dim2];

            float[] retval = new float[m_rmdat.Count * 2];
            for (int i = 0; i < m_rmdat.Count; i++)
            {
                int cpos = i * 2;
                retval[cpos    ] = (float)m_rmdat[i][dim1];
                retval[cpos + 1] = (float)m_rmdat[i][dim2];
            }

            m_pcache[dim1][dim2] = retval;

            return retval;
        }

        /// Liefert die Anzahl an Dimensionen
        /// <returns>Anzahl aller Dimensionen</returns>
        public int getDimensions()
        {
            return m_ncols;
        }

        /// Dimensionsnamen
        /// <returns>Array mit allen Dimensionsbezeichnern</returns>
        public String[] getDimensionTitles()
        {
            return m_dimtitles;
        }

        /// Liefert das Maximum über alle Dimensionen
        /// <returns>Maximum</returns>
        public float getGlobalMaximum()
        {
            return (float)m_yglobalmaximum;
        }

        /// Liefert das Maximum der ausgewählten Dimensionen
        /// <param name="dims">Dimensionen, in denen nach dem Maximum gesucht werden soll</param>
        /// <returns>Maximum</returns>
        public float getGlobalMaximum(int[] dims)
        { 
            double max = double.NegativeInfinity;
            for (int i = 0; i < dims.Length; i++)
                max = Math.Max(max, m_ymaxima[dims[i]]);
            return (float)max;
        }

        /// Liefert das Maximum einer einzelnen Dimension
        /// <param name="dim">Index der Dimension, von der das Maximum herausgesucht werden soll</param>
        /// <returns>Maximum</returns>
        public float getMaximum(int dim)
        {
            return (float)m_ymaxima[dim];
        }

        /// Liefert das Minimum über alle Dimensionen
        /// <returns>Minimum</returns>
        public float getGlobalMinimum()
        {
            return (float)m_yglobalminimum;
        }

        /// Liefert das Minimum der ausgewählten Dimensionen
        /// <param name="dims">Dimensionen, in denen nach dem Minimum gesucht werden soll</param>
        /// <returns>Minimum</returns>
        public float getGlobalMinimum(int[] dims)
        {
            double min = double.PositiveInfinity;
            for (int i = 0; i < dims.Length; i++)
                min = Math.Min(min, m_yminima[dims[i]]);
            return (float)min;
        }

        /// Liefert das Minimum einer einzelnen Dimension
        /// <param name="dim">Index der Dimension, von der das Minimum herausgesucht werden soll</param>
        /// <returns>Minimum</returns>
        public float getMinimum(int dim)
        {
            return (float)m_yminima[dim];
        }
    }
}
