﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Helper-Klasse welche Vertex Attribut-Daten wie Name, Länge und Offset bezüglich eines VBOs speichert
    class PrimitiveDataInfo
    {
        private class AttributeInfo
        {
            public string Name;
            public int Length;
            public int Offset;

            public AttributeInfo(string _name, int _length, int _offset)
            {
                Name = _name;
                Length = _length;
                Offset = _offset;
            }
        }

        private List<AttributeInfo> mAttributeList;
        private BeginMode mPrimitiveMode;
        private int mTotalOffsetOverride;

        /// Konstruktor
        /// <param name="_primitiveMode">die Art der Primitive</param>
        public PrimitiveDataInfo(BeginMode _primitiveMode)
        {
            mTotalOffsetOverride = 0;
            mPrimitiveMode = _primitiveMode;
            mAttributeList = new List<AttributeInfo>();
        }

        /// Fügt ein Attribut hinzu.
        /// <param name="_name">der Name</param>
        /// <param name="_length">die Länge</param>
        /// <param name="_offset">der Offset</param>
        public void appendAttribute(string _name, int _length, int _offset)
        {
            mAttributeList.Add(new AttributeInfo(_name, _length, _offset));
        }

        /// Überschreibt den totalen offset
        /// <param name="_offset">der Offset</param>
        public void setTotalOffset(int _offset)
        {
            if (_offset < 1)
                throw new Exception("Offset < 1");

            mTotalOffsetOverride = _offset;
        }

        /// Liefert den Attribut Namen über den Index zurück.
        /// <param name="_i">der Index</param>
        /// <returns>der Name</returns>
        public string getNameByIndex(int _i)
        {
            return mAttributeList.ElementAt(_i).Name;
        }

        /// Liefert die Attribut Länge über den Index zurück.
        /// <param name="_i">der Index</param>
        /// <returns>die Länge</returns>
        public int getLengthByIndex(int _i)
        {
            return mAttributeList.ElementAt(_i).Length;
        }

        /// Liefert den Attribut Offset über den Index zurück.
        /// <param name="_i">der Index</param>
        /// <returns>der Offset</returns>
        public int getOffsetByIndex(int _i)
        {
            return mAttributeList.ElementAt(_i).Offset;
        }

        /// Liefert die Arte der Primitive zurück.
        /// <returns>Arte der Primitive</returns>
        public BeginMode getPrimitiveMode()
        {
            return mPrimitiveMode;
        }

        /// Liefert den totalen Offset zurück.
        /// <returns>den totalen Offset</returns>
        public int getTotalOffset()
        {
            if (mTotalOffsetOverride > 0)
                return mTotalOffsetOverride;

            // assuming that the last attribute in the list refers to the last attribute in the data
            return mAttributeList.Last().Offset + mAttributeList.Last().Length;
        }

        /// /// Liefert die Anzahl der Attribute zurück.
        /// <returns>die Anzahl der Attribute</returns>
        public int getAttributeCount()
        {
            return mAttributeList.Count;
        }

        /// Liefert ein Array der Attribut-Namen zurück.
        /// <returns>Array der Attribut-Namen</returns>
        public string[] getNames()
        {
            string[] names = new string[mAttributeList.Count];

            for (int i = 0; i < mAttributeList.Count; i++)
                names[i] = mAttributeList.ElementAt(i).Name;

            return names;
        }
    }
}
