﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Eine Utility Klasse mit nützlichen Methoden
    static class Utility
    {
        /// Läd eine Datei in einen String
        /// <param name="filename">der Datei-Pfad</param>
        /// <returns>der String</returns>
        public static String fileToStr(String filename)
        {
            StreamReader sr = new StreamReader(filename);
            String filecont = "";
            try
            {
                String line;
                do
                {
                    line = sr.ReadLine();
                    filecont += (line + "\n");
                } while (line != null);
            }
            finally
            {
                sr.Close();
            }
            return filecont;
        }

        /// Prüft auf OpenGL Error und gibt diesen gegebenenfalls aus
        /// <param name="_location">Beschreibung der Aufruf-Stelle</param>
        public static void GLErrorCheck(string _location = null)
        {
            ErrorCode error = GL.GetError();

            if (error == ErrorCode.NoError)
                return;

            Console.Write("GL Error: " + error.ToString());            
            if (_location == null) Console.WriteLine();
            else Console.WriteLine(" @ " + _location);
        }
        
        /// Liefert ein zufälliges float Array zurück
        /// <param name="_size">die Array Größe</param>
        /// <param name="_intervalMin">Interval-Start</param>
        /// <param name="_intervalMax">Interval-Ende</param>
        /// <returns>zufälliges float Array</returns>
        public static float[] getRandomFloatArray(int _size, float _intervalMin = 0, float _intervalMax = 1)
        {
            Random randGen = new Random();
            
            float[] random = new float[_size];

            for (int i = 0; i < random.Length; i++)
            {
                random[i] = (float)(randGen.NextDouble() * (_intervalMax - _intervalMin) + _intervalMin);
            }

            return random;
        }

        /// Liefert das float Array eines NDC-ScreenQuads mit UV-Koordinaten zurück (für Upload in ein VBO)
        /// <param name="_halfSideLength">die halbe Seitenlänge</param>
        /// <returns>float Array eines NDC-ScreenQuads</returns>
        public static float[] getNdcQuadWithUV(float _halfSideLength = 1.0f)
        {
            float[] quadCoord = new float[24];
            quadCoord[0] = _halfSideLength; quadCoord[1] = _halfSideLength;
            quadCoord[2] = 1.0f; quadCoord[3] = 1.0f;

            quadCoord[4] = -_halfSideLength; quadCoord[5] = _halfSideLength;
            quadCoord[6] = 0.0f; quadCoord[7] = 1.0f;

            quadCoord[8] = -_halfSideLength; quadCoord[9] = -_halfSideLength;
            quadCoord[10] = 0.0f; quadCoord[11] = 0.0f;

            quadCoord[12] = _halfSideLength; quadCoord[13] = _halfSideLength;
            quadCoord[14] = 1.0f; quadCoord[15] = 1.0f;

            quadCoord[16] = -_halfSideLength; quadCoord[17] = -_halfSideLength;
            quadCoord[18] = 0.0f; quadCoord[19] = 0.0f;

            quadCoord[20] = _halfSideLength; quadCoord[21] = -_halfSideLength;
            quadCoord[22] = 1.0f; quadCoord[23] = 0.0f;

            return quadCoord;
        }
    }
}
