#version 150
#extension GL_EXT_gpu_shader4 : enable

uniform sampler2D noiseTex;
uniform sampler2D densityTex;
uniform int sampleCount;

uniform int axisCount;
uniform int currentAxisIndex;

layout(points) in;
layout(line_strip, max_vertices = 2) out;

out vec4 vertexColor;

vec4 convertToFlatNDC(vec2 _pos)
{
	return vec4(_pos * 2.0 - vec2(1.0), 0.0, 1.0);
}

void main()
{
	// calculate texel-fetch position using gl_PrimitiveIDIn
	
	float sampleStep = 1.0 / sampleCount;
	float halfSampleStep = sampleStep * 0.5;
	
	// sample grid position
	int y = gl_PrimitiveIDIn / sampleCount;
	int x = gl_PrimitiveIDIn - y * sampleCount;

	// actual noise texel fetch (UV) coordinates
	float tx = halfSampleStep + sampleStep * x;
	float ty = halfSampleStep + sampleStep * y;
	vec2 noiseUvPos = vec2(tx, ty);

	// sample noise value
	vec4 noise = texture2D(noiseTex, noiseUvPos);
	
	// calc jitter for regular sampling
	vec2 jitter = (noise.xy - vec2(0.5)) * sampleStep;
	
	vec2 uvPos = noiseUvPos + jitter;

	// flip sampling pattern
	if(mod(currentAxisIndex, 2) == 1) uvPos = uvPos.yx;

	// DEACTIVATE RANDOM SAMPLING
	//uvPos = noiseUvPos;

	// read density value
	vec4 density = texture2D(densityTex, uvPos);
	if(density.a == 0) return;
	density.a = 1; /* remember: alpha is blended nonsense */

	// calculate line vertices
	
	float xStep = 1.0 / float(axisCount - 1);
	float x1 = currentAxisIndex * xStep;
	float x2 = (currentAxisIndex + 1) * xStep;
	
	// calculate y values from point-line dualism	
	vec2 pos1 = vec2(x1, uvPos.x);
	vec2 pos2 = vec2(x2, uvPos.y);

	if(density.a == 0) return;

	// generate line primitive
	gl_Position = convertToFlatNDC(pos1);
	vertexColor = density; 
	EmitVertex();
		
	gl_Position = convertToFlatNDC(pos2);
	vertexColor = density;
	EmitVertex();

	EndPrimitive();
}