﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace HW09CPC
{
    /// Steuerelement: Konfiguriert die Anzeige der Parallelkoordinatenansicht
    public partial class ucSteering : UserControl
    {
        private PcCvsDataset m_ds = null;

        public delegate void DimensionEventHandler(Object sender, EventArgs e);
        /// Wird ausgelöst, sobald die Auswahl oder die Reihenfolge der Dimensionen geändert wird
        public event DimensionEventHandler DimensionOrderChanged;

        public delegate void RenderModeEventHandler(Object sender, RenderModeEventargs e);
        /// Wird ausgelöst, wenn der Benutzer einen anderen Rendermodus auswählt
        public event RenderModeEventHandler RenderModeChanged;

        /// Mögliche Rendermodi
        public enum RenderModeEnum {Simple, Colormap, Continuous};

        public delegate void ColormapEventHandler(Object sender, ColormapEventargs e);
        /// Wird ausgelöst, wenn der Benutzer eine andere Colormap auswählt
        public event ColormapEventHandler ColormapChanged;

        public delegate void ContinuousParsEventHandler(Object sender, ContinuousParsEventargs e);
        /// Wird ausgelöst, wenn sich Parameter bezüglich der kontinuierlichen Darstellung ändern
        public event ContinuousParsEventHandler ContinuousParsChanged;

        public delegate void TransformParsEventHandler(Object sender, TransformParsEventargs e);
        /// Wird ausgelöst, wenn der Benutzer seine Einstellung des "local transform" ändert
        public event TransformParsEventHandler TransformParsChanged;

        /// Mögliche Colormaps
        public enum ColormapEnum { Gray, Heat, Rainbow };

        /// Erstellt ein neues Steuerelement
        public ucSteering()
        {
            InitializeComponent();
        }

        private void ucSteering_Load(object sender, EventArgs e)
        {
            cmbMode.SelectedIndex = 0;
            cmbMap.SelectedIndex = 0;
            setDimensions(null);
        }

        /// Legt die Anzahl/Namen/Reihenfolge der Dimensionen fest
        /// <param name="ds">Der geladene Datensatz</param>
        public void setDimensions(PcCvsDataset ds)
        {
            m_ds = ds;
            int dims = (ds != null) ? ds.getDimensions() : 0;

            clbDimensions.Items.Clear();
            for (int i = 0; i < dims; i++)
            {
                clbDimensions.Items.Add(new DimListItem(i, ds.getDimensionTitles()[i]));
                clbDimensions.SetItemChecked(i, true);
            }

            bool isPos = (dims > 0);
            clbDimensions.Enabled = isPos;
            cmbMode.Enabled = isPos;

            checkForm(); /* kann evt. wieder Elemente aktivieren, nachher wieder entsprechen deaktivieren. */

            cmbMap.Enabled = cmbMap.Enabled && isPos;
            chkBilinear.Enabled = chkBilinear.Enabled && isPos;

        }

        private void checkForm()
        {
            bool hasmap = cmbMode.SelectedIndex == 1 || cmbMode.SelectedIndex == 2;
            bool doesint = cmbMode.SelectedIndex == 2;

            cmbMap.Enabled = hasmap;
            nupSWidth.Enabled = doesint;
            nupSpHorizontal.Enabled = doesint;
            chkBilinear.Enabled = doesint;

        }

        private void cmbMode_SelectedIndexChanged(object sender, EventArgs e)
        {
            checkForm();
            OnRenderModeChanged();
        }

        private void clbDimensions_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if(e.NewValue == e.CurrentValue) return;
            OnDimensionOrderChanged();
        }

        /// Liefert die aktuelle Dimensionsauswahl zurück
        /// <returns>Generiertes Array mit momentan ausgewählten Dimensionen in korrekter Reihenfolge</returns>
        public int[] getOrderedDimSelection()
        {
            int[] retval = new int[clbDimensions.CheckedItems.Count];
            for (int i = 0; i < clbDimensions.CheckedItems.Count; i++ )
                retval[i] = ((DimListItem)clbDimensions.CheckedItems[i]).GetDimension();

            return retval;
        }

        private void OnDimensionOrderChanged()
        {
            if (DimensionOrderChanged != null)
                DimensionOrderChanged(this, new EventArgs());
        }

        private void OnRenderModeChanged()
        {
            if (RenderModeChanged != null)
            {
                switch (cmbMode.SelectedIndex)
                {
                    case 0:
                        RenderModeChanged(this, new RenderModeEventargs(RenderModeEnum.Simple));
                        break;
                    case 1:
                        RenderModeChanged(this, new RenderModeEventargs(RenderModeEnum.Colormap));
                        break;
                    case 2:
                        RenderModeChanged(this, new RenderModeEventargs(RenderModeEnum.Continuous));
                        break;
                    default:
                        throw new InvalidOperationException("Es muss ein Rendermode aktiv sein");
                }
            }
        }

        private void OnColormapChanged()
        {
            if (ColormapChanged != null)
            {
                switch (cmbMap.SelectedIndex)
                { 
                    case 0:
                        ColormapChanged(this, new ColormapEventargs(ColormapEnum.Gray));
                        break;
                    case 1:
                        ColormapChanged(this, new ColormapEventargs(ColormapEnum.Heat));
                        break;
                    case 2:
                        ColormapChanged(this, new ColormapEventargs(ColormapEnum.Rainbow));
                        break;
                    default:
                        throw new InvalidOperationException("Es muss eine Colormap aktiv sein");
                }
            }
        }

        private void OnContinuousParsChanged()
        {
            if (ContinuousParsChanged != null)
            {
                ContinuousParsEventargs cpe = new ContinuousParsEventargs((int)nupSWidth.Value, (int)nupSpHorizontal.Value, chkBilinear.Checked);
                ContinuousParsChanged(this, cpe);
            }
        }

        private void OnTransformParsChanged()
        {
            if (TransformParsChanged != null)
            {
                TransformParsChanged(this, new TransformParsEventargs(chkY01.Checked));
            }
        }

        private void btnUp_Click(object sender, EventArgs e)
        {
            int sind = clbDimensions.SelectedIndex;
            if (sind == -1 || sind == 0) return;
            
            DimListItem dli = (DimListItem)clbDimensions.Items[sind];
            bool waschecked = clbDimensions.CheckedItems.Contains(dli);

            clbDimensions.Items.RemoveAt(sind);
            clbDimensions.Items.Insert(sind - 1, dli);
            clbDimensions.SetItemChecked(sind - 1, waschecked);

            clbDimensions.SetSelected(sind - 1, true);
            
        }

        private void btnDown_Click(object sender, EventArgs e)
        {
            int sind = clbDimensions.SelectedIndex;
            if (sind == -1 || sind == clbDimensions.Items.Count - 1) return;

            DimListItem dli = (DimListItem)clbDimensions.Items[sind];
            bool waschecked = clbDimensions.CheckedItems.Contains(dli);

            clbDimensions.Items.RemoveAt(sind);
            clbDimensions.Items.Insert(sind + 1, dli);
            clbDimensions.SetItemChecked(sind + 1, waschecked);

            clbDimensions.SetSelected(sind + 1, true);
        }

        private void cmbMap_SelectedIndexChanged(object sender, EventArgs e)
        {
            OnColormapChanged();
        }

        private class DimListItem
        {
            private int m_dim;
            private string m_title;

            public DimListItem(int dimension, string title)
            {
                m_dim = dimension;
                m_title = title;
            }

            public int GetDimension()
            {
                return m_dim;
            }

            public override String ToString()
            {
                return m_title;
            }

        }

        /// Event-Parameter für das Ändern des Rendermodus
        public class RenderModeEventargs : EventArgs
        {
            private RenderModeEnum m_rm;

            public RenderModeEventargs(RenderModeEnum rm)
            {
                m_rm = rm;
            }

            public RenderModeEnum getRenderMode()
            {
                return m_rm;
            }
        }

        /// Event-Parameter für das Ändern der Colormap
        public class ColormapEventargs : EventArgs
        {
            private ColormapEnum m_cm;

            public ColormapEventargs(ColormapEnum cm)
            {
                m_cm = cm;
            }

            public ColormapEnum getRenderMode()
            {
                return m_cm;
            }
        }

        /// Event-Parameter für das Ändern der kontinuierlichen Darstellung
        public class ContinuousParsEventargs : EventArgs
        {
            private int m_sigmawidth, m_samplingprec;
            private bool m_usebilinear;

            public ContinuousParsEventargs(int SigmaWidth, int SamplingPrec, bool useBilinear)
            {
                m_sigmawidth = SigmaWidth;
                m_samplingprec = SamplingPrec;
                m_usebilinear = useBilinear;
            }

            public int getSigmaWidth()
            {
                return m_sigmawidth;
            }

            public int getSamplingPrec()
            {
                return m_samplingprec;
            }

            public bool getUseBilinear()
            {
                return m_usebilinear;
            }
        }

        /// Event-Parameter für das Ändern der Einstellung "local transform"
        public class TransformParsEventargs : EventArgs
        {
            private bool m_localtransform;

            public TransformParsEventargs(bool LocalTransform)
            {
                m_localtransform = LocalTransform;
            }

            public bool getLocalTransform()
            {
                return m_localtransform;
            }
        }

        private void nupSWidth_ValueChanged(object sender, EventArgs e)
        {
            OnContinuousParsChanged();
        }

        private void nupSpHorizontal_ValueChanged(object sender, EventArgs e)
        {
            OnContinuousParsChanged();
        }

        private void chkBilinear_CheckedChanged(object sender, EventArgs e)
        {
            OnContinuousParsChanged();
        }

        private void chkY01_CheckedChanged(object sender, EventArgs e)
        {
            OnTransformParsChanged();
        }
    }
}
